/* Copyright (C) 2022-2023 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergei Zharko [committer] */

/// @file   CbmL1DetectorID.h
/// @brief  Implementation of L1DetectorID enum class for CBM
/// @author S.Zharko
/// @since  01.12.2022

#ifndef CbmL1DetectorID_h
#define CbmL1DetectorID_h 1

#include <string>

#include "L1EArray.h"

/// Enumeration for the detector subsystems used in L1 tracking
/// It is important for the subsystems to be specified in the actual order. The order is used
/// for the L1Station array filling.
/// Note: L1DetectorID has a forward declaration in L1InitManager.h and L1BaseStationInfo.h
/// TODO: Probably, we need to move everything into a single CbmCaConst.h header
/// NOTE: The enumeration must not contain jumps in ordering and the first entry must be equal 0
enum class L1DetectorID
{
  kMvd = 0,
  kSts,
  kMuch,
  kTrd,
  kTof,
  kEND  ///< End of enumeration
};

/// @brief  Alias to array, indexed by L1DetectorID enum
/// @note   To be used only in CBM-specific code
template<typename T>
using CbmCaDetIdArr_t = L1EArray<L1DetectorID, T>;

/// @brief  Array of types, indexed by L1DetectorID enum
///
/// The array of types allows to treat different types of detector in a uniform manner
/// Example:
///   using HitTypes_t = CbmCaDetIdTypeArr_t<CbmMvdHit, CbmStsHit, CbmMuchPixelHit, CbmTrdHit, CbmTofHit>;
///   ...
///   HitTypes_t::at<L1DetectorID::kSts> hit;  // Sts hit
template<class... Types>
struct CbmCaDetIdTypeArr_t {
  template<L1DetectorID DetID>
  using at                          = std::tuple_element_t<static_cast<std::size_t>(DetID), std::tuple<Types...>>;
  static constexpr std::size_t size = sizeof...(Types);
};

/// *************************************************
/// **    Detector-dependent common definitions    **
/// *************************************************

class CbmMvdPoint;
class CbmStsPoint;
class CbmMuchPoint;
class CbmTrdPoint;
class CbmTofPoint;

class CbmMvdHit;
class CbmStsHit;
class CbmMuchPixelHit;
class CbmTrdHit;
class CbmTofHit;

namespace cbm::ca
{
  /// @brief Names of detector subsystems
  /// @note  These names are used for data branches IO, thus any modification can lead to the
  ///        read-out corruption.
  constexpr CbmCaDetIdArr_t<const char*> kDetName = {{"MVD", "STS", "MUCH", "TRD", "TOF"}};


  /// @brief Types of MC point objects for each detector
  using PointTypes_t = CbmCaDetIdTypeArr_t<CbmMvdPoint, CbmStsPoint, CbmMuchPoint, CbmTrdPoint, CbmTofPoint>;

  /// @brief Types of hit objects for each detector
  using HitTypes_t = CbmCaDetIdTypeArr_t<CbmMvdHit, CbmStsHit, CbmMuchPixelHit, CbmTrdHit, CbmTofHit>;

}  // namespace cbm::ca

/// @brief Enumeration for different tracking running modes
enum class ECbmCaTrackingMode
{
  kSTS,  ///< Local tracking in CBM (STS + MVD), results stored to the StsTrack branch
  kMCBM  ///< Global tracking in mCBM (STS, MuCh, TRD, TOF), results stored to GlobalTrack branch
};


#endif  // CbmL1DetectorID_h
