/* Copyright (C) 2023 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergey Gorbunov [committer] */


/// \file   CaToolsMaterialHelper.h
/// \brief  A helper class to create material maps (header)
/// \since  29.06.2023
/// \author S.Gorbunov


#ifndef CaToolsMaterialHelper_H
#define CaToolsMaterialHelper_H

#include "Rtypes.h"
#include "TObject.h"

#include "L1Material.h"

class TGeoNavigator;

namespace ca
{
  namespace tools
  {
    /// Class ca::tools::Debugger helps to debug CA tracking
    ///
    ///
    /// class to create L1Material material maps form the ROOT geometry
    ///
    class MaterialHelper : public TObject {
    public:
      MaterialHelper();
      ~MaterialHelper();

      ///
      /// project rays radially from the targetZ througth the XY-bins at a reference z.
      ///
      void SetDoRadialProjection(double targetZ)
      {
        fDoRadialProjection = true;
        fTargetZ            = targetZ;
      }

      ///
      /// project rays horisontally along the Z axis (default)
      ///
      void SetDoHorisontalProjection() { fDoRadialProjection = false; }

      ///
      /// shoot nRaysDim * nRaysDim rays for each bin in the map
      ///
      void SetNraysPerDim(int nRaysDim) { fNraysBinPerDim = (nRaysDim > 0) ? nRaysDim : 1; }


      /// generate a material map, collecting the material between [zMin, zMax]
      /// with radial rays from (0,0,targetZ) througth the XY-bins at z == zRef.
      ///
      /// It creates a map with [nBinsDim x nBinsDim] bins, of a size of [+-xyMax, +-xyMax]
      /// shooting fNraysBinPerDim x fNraysBinPerDim through each bin
      ///
      /// The calculated radiation thickness is a projection of the rad.thickness along the ray onto the Z axis.
      /// RadThick = sum of (dZ / radLength) over ray trajectory pieces
      ///
      /// When doRadialProjection==false the rays are shoot horisontally along the Z axis
      ///
      L1Material GenerateMaterialMap(double zRef, double zMin, double zMax, double xyMax, int nBinsDim);

      void SetSafeMaterialInitialization(bool val = true) { fDoSafeInitialization = val; }

    private:
      /// initialise the necessary amount of threads in TGeoManager
      void InitThreads();

      /// Clean up TGeoManager: threadIds, create a defailt navigator
      void CleanUpThreads();

      /// Get navigator for current thread, create it if it doesnt exist.
      /// Produce an error when anything goes wrong
      TGeoNavigator* GetCurrentNavigator(int iThread);

    private:
      int fNthreadsOld {0};                        // number of threads in TGeoManager before the helper was created
      int fNthreads {0};                           // number of threads
      bool fDoRadialProjection {false};            // project rays horisontally along the Z axis (special mode)
      int fNraysBinPerDim {3};                     // shoot fNraysBinPerDim * fNraysBinPerDim rays in each map bin
      double fTargetZ {0.};                        // z of the target for the radial projection
      std::vector<TGeoNavigator*> fNavigators {};  // list of created navigators
      bool fDoSafeInitialization {false};          // perform slow but safe initialization
                                                   // to get around the crashes in TGeoVoxelFinder
      ClassDef(MaterialHelper, 0);
    };

  }  // namespace tools
}  // namespace ca


#endif
