/* Copyright (C) 2023 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergei Zharko [committer] */

/// \file   CbmCaInputQaTrd.cxx
/// \date   13.01.2023
/// \brief  QA-task for CA tracking input from TRD detector (implementation)
/// \author S.Zharko <s.zharko@gsi.de>

#include "CbmCaInputQaTrd.h"

#include "CbmAddress.h"
#include "CbmMCDataArray.h"
#include "CbmMCEventList.h"
#include "CbmMCTrack.h"
#include "CbmMatch.h"
#include "CbmQaCanvas.h"
#include "CbmQaEff.h"
#include "CbmTimeSlice.h"
#include "CbmTrdHit.h"
#include "CbmTrdPoint.h"
#include "CbmTrdTrackingInterface.h"

#include "FairRootManager.h"
#include "Logger.h"

#include "TBox.h"
#include "TClonesArray.h"
#include "TEllipse.h"
#include "TF1.h"
#include "TFormula.h"
#include "TGraphAsymmErrors.h"
#include "TH1F.h"
#include "TH2F.h"
#include "TMath.h"
#include "TStyle.h"

#include <algorithm>
#include <fstream>
#include <iomanip>
#include <numeric>

#include "L1Constants.h"

ClassImp(CbmCaInputQaTrd);

namespace phys = L1Constants::phys;  // from L1Constants.h

// ---------------------------------------------------------------------------------------------------------------------
//
CbmCaInputQaTrd::CbmCaInputQaTrd(int verbose, bool isMCUsed) : CbmQaTask("CbmCaInputQaTrd", verbose, isMCUsed) {}

// ---------------------------------------------------------------------------------------------------------------------
//
bool CbmCaInputQaTrd::Check()
{
  bool res = true;

  int nSt = fpDetInterface->GetNtrackingStations();


  // **************************************************************
  // ** Basic checks, available both for real and simulated data **
  // **************************************************************

  // ----- Checks for mismatches in the ordering of the stations
  //
  std::vector<double> vStationPos(nSt, 0.);
  for (int iSt = 0; iSt < nSt; ++iSt) {
    vStationPos[iSt] = fpDetInterface->GetZ(iSt);
  }

  if (!std::is_sorted(vStationPos.cbegin(), vStationPos.cend(), [](int l, int r) { return l <= r; })) {
    if (fVerbose > 0) {
      LOG(error) << fName << ": stations are ordered improperly along the beam axis:";
      for (auto zPos : vStationPos) {
        LOG(error) << "\t- " << zPos;
      }
    }
    res = false;
  }

  // ----- Checks for mismatch between station and hit z positions
  //   The purpose of this block is to be ensured, that hits belong to the correct tracking station. For each tracking
  // station a unified position z-coordinate is defined, which generally differs from the corresponding positions of
  // reconstructed hits. This is due to non-regular position along the beam axis for each detector sensor. Thus, the
  // positions of the hits along z-axis are distributed relatively to the position of the station in some range.
  // If hits goes out this range, it can signal about a mismatch between hits and geometry. For limits of the range
  // one should select large enough values to cover the difference described above and in the same time small enough
  // to avoid overlaps with the neighboring stations.
  //   For each station, a distribution of z_{hit} - z_{st} is integrated over the defined range and scaled by the
  // total number of entries to the distribution. If this value is smaller then unity, then some of the hits belong to
  // another station.
  //
  for (int iSt = 0; iSt < nSt; ++iSt) {
    int nHits = fvph_hit_station_delta_z[iSt]->GetEntries();
    if (!nHits) {
      LOG_IF(error, fVerbose > 0) << fName << ": station " << iSt << " does not have hits";
      res = false;
      continue;
    }
    int iBinMin = fvph_hit_station_delta_z[iSt]->FindBin(-fMaxDiffZStHit);
    int iBinMax = fvph_hit_station_delta_z[iSt]->FindBin(+fMaxDiffZStHit);

    if (fvph_hit_station_delta_z[iSt]->Integral(iBinMin, iBinMax) < nHits) {
      LOG_IF(error, fVerbose > 0) << fName << ": station " << iSt << " has mismatches in hit z-positions";
      res = false;
    }
  }


  // *******************************************************
  // ** Additional checks, if MC information is available **
  // *******************************************************

  if (IsMCUsed()) {
    // ----- Check efficiencies
    // Error is raised, if any station has integrated efficiency lower then a selected threshold.
    // Integrated efficiency is estimated with fitting the efficiency curve for hit distances (r) with a uniform
    // distribution in the range from kEffRangeMin to kEffRangeMax, where the efficiency is assigned to be constant
    //
    // Fit efficiency curves
    LOG(info) << "-- Hit efficiency integrated over hit distance from station center";
    LOG(info) << "\tintegration range: [" << fEffRange[0] << ", " << fEffRange[1] << "] cm";

    auto* pEffTable = MakeQaObject<CbmQaTable>("eff_table", "Efficiency table", nSt, 2);
    pEffTable->SetNamesOfCols({"Station ID", "Efficiency"});
    pEffTable->SetColWidth(20);

    for (int iSt = 0; iSt < nSt; ++iSt) {
      auto eff = fvpe_reco_eff_vs_r[iSt]->GetMean(2);
      pEffTable->SetCell(iSt, 0, iSt);
      pEffTable->SetCell(iSt, 1, eff);
      res = CheckRange("Hit finder efficiency in station " + std::to_string(iSt), eff, fEffThrsh, 1.000);
    }

    LOG(info) << '\n' << pEffTable->ToString(3);

    // ----- Checks for residuals
    // Check hits for potential biases from central values

    // Function to fit a residual distribution, returns a structure
    auto FitResidualsAndGetMean = [&](TH1* pHist) {
      auto fit        = TF1("fitres", "gausn");
      double statMean = pHist->GetMean();
      double statSigm = pHist->GetStdDev();
      fit.SetParameters(100., statMean, statSigm);
      pHist->Fit("fitres", "MQ");
      pHist->Fit("fitres", "MQ");
      pHist->Fit("fitres", "MQ");
      // NOTE: Several fit procedures are made to avoid empty fit results
      std::array<double, 3> result;
      result[0] = fit.GetParameter(1);
      result[1] = -fit.GetParameter(2) * fResMeanThrsh;
      result[2] = +fit.GetParameter(2) * fResMeanThrsh;
      return result;
    };

    auto* pResidualsTable =
      MakeQaObject<CbmQaTable>("residuals_mean", "Residual mean values in different stations", nSt, 4);
    pResidualsTable->SetNamesOfCols({"Station ID", "Residual(x) [cm]", "Residual(y) [cm]", "Residual(t) [ns]"});
    pResidualsTable->SetColWidth(20);

    // Fill residuals fit results and check boundaries
    for (int iSt = 0; iSt < nSt; ++iSt) {
      auto fitX = FitResidualsAndGetMean(fvph_res_x[iSt]);
      auto fitY = FitResidualsAndGetMean(fvph_res_y[iSt]);
      auto fitT = FitResidualsAndGetMean(fvph_res_t[iSt]);
      res = CheckRange("Mean of x residual [cm] in station " + std::to_string(iSt), fitX[0], fitX[1], fitX[2]) && res;
      res = CheckRange("Mean of y residual [cm] in station " + std::to_string(iSt), fitY[0], fitY[1], fitY[2]) && res;
      res = CheckRange("Mean of t residual [ns] in station " + std::to_string(iSt), fitT[0], fitT[1], fitT[2]) && res;
      pResidualsTable->SetCell(iSt, 0, iSt);
      pResidualsTable->SetCell(iSt, 1, fitX[0]);
      pResidualsTable->SetCell(iSt, 2, fitX[1]);
      pResidualsTable->SetCell(iSt, 3, fitX[2]);
    }

    LOG(info) << '\n' << pResidualsTable->ToString(8);

    // ----- Checks for pulls
    // For the hit pull is defined as a ration of the difference between hit and MC-point position or time component
    // values and an error of the hit position (or time) component, calculated in the same z-positions. In ideal case,
    // when the resolutions are well determined for detecting stations, the pull distributions should have a RMS equal
    // to unity. Here we allow a RMS of the pull distributions to be varied in a predefined range. If the RMS runs out
    // this range, QA task fails.
    // TODO: Add checks for center

    // Fit pull distributions

    // ** Kaniadakis Gaussian distribution, gives smaller chi2 / NDF
    //if (!gROOT->FindObject("Expk")) {
    //  new TFormula("Expk", "TMath::Power(TMath::Sqrt(1 + x[1] * x[1] * x[0] * x[0]) + x[0] * x[1], 1./x[1])");
    //}
    //TF1* pPullFit = new TF1("pullFitGausn", "[0] * Expk(-[2] * (x - [1]) * (x - [1]), [3])", -10., 10.);

    auto FitPullsAndGetSigma = [&](TH1* pHist) {
      auto fit = TF1("fitpull", "gausn(0)");
      fit.SetParameters(100, 0.001, 1.000);
      pHist->Fit("fitpull", "Q");
      return fit.GetParameter(2);
    };

    auto* pPullsTable = MakeQaObject<CbmQaTable>("pulls_rms", "Pulls std. dev. values in different stations", nSt, 4);
    pPullsTable->SetNamesOfCols({"Station ID", "Pull(x) sigm", "Pull(y) sigm", "Pull(z) sigm"});
    pPullsTable->SetColWidth(20);

    for (int iSt = 0; iSt < nSt; ++iSt) {
      double rmsPullX = FitPullsAndGetSigma(fvph_pull_x[iSt]);
      double rmsPullY = FitPullsAndGetSigma(fvph_pull_y[iSt]);
      double rmsPullT = FitPullsAndGetSigma(fvph_pull_t[iSt]);

      double rmsPullHi = 1. + fPullWidthThrsh;
      double rmsPullLo = 1. - fPullWidthThrsh;

      res = CheckRange("Rms for x pull in station " + std::to_string(iSt), rmsPullX, rmsPullLo, rmsPullHi) && res;
      res = CheckRange("Rms for y pull in station " + std::to_string(iSt), rmsPullY, rmsPullLo, rmsPullHi) && res;
      res = CheckRange("Rms for t pull in station " + std::to_string(iSt), rmsPullT, rmsPullLo, rmsPullHi) && res;
      pPullsTable->SetCell(iSt, 0, iSt);
      pPullsTable->SetCell(iSt, 1, rmsPullX);
      pPullsTable->SetCell(iSt, 2, rmsPullY);
      pPullsTable->SetCell(iSt, 3, rmsPullT);
    }

    LOG(info) << '\n' << pPullsTable->ToString(3);
  }

  return res;
}

// ---------------------------------------------------------------------------------------------------------------------
//
void CbmCaInputQaTrd::DeInit()
{
  // Vectors with pointers to histograms
  fvph_hit_ypos_vs_xpos.clear();
  fvph_hit_xpos_vs_zpos.clear();
  fvph_hit_ypos_vs_zpos.clear();

  fvph_hit_station_delta_z.clear();

  fvph_hit_dx.clear();
  fvph_hit_dy.clear();
  fvph_hit_dt.clear();

  fvph_n_points_per_hit.clear();

  fvph_point_ypos_vs_xpos.clear();
  fvph_point_xpos_vs_zpos.clear();
  fvph_point_ypos_vs_zpos.clear();

  fvph_point_hit_delta_z.clear();

  fvph_res_x.clear();
  fvph_res_y.clear();
  fvph_res_t.clear();

  fvph_pull_x.clear();
  fvph_pull_y.clear();
  fvph_pull_t.clear();

  fvph_res_x_vs_x.clear();
  fvph_res_y_vs_y.clear();
  fvph_res_t_vs_t.clear();

  fvph_pull_x_vs_x.clear();
  fvph_pull_y_vs_y.clear();
  fvph_pull_t_vs_t.clear();

  fvpe_reco_eff_vs_xy.clear();
  fvpe_reco_eff_vs_r.clear();
}

// ---------------------------------------------------------------------------------------------------------------------
//
void CbmCaInputQaTrd::FillHistograms()
{
  int nSt       = fpDetInterface->GetNtrackingStations();
  int nHits     = fpHits->GetEntriesFast();
  int nMCevents = (IsMCUsed()) ? fpMCEventList->GetNofEvents() : -1;

  std::vector<std::vector<int>> vNofHitsPerPoint;  // Number of hits per MC point in different MC events

  if (IsMCUsed()) {
    vNofHitsPerPoint.resize(nMCevents);
    for (int iE = 0; iE < nMCevents; ++iE) {
      int iFile   = fpMCEventList->GetFileIdByIndex(iE);
      int iEvent  = fpMCEventList->GetEventIdByIndex(iE);
      int nPoints = fpMCPoints->Size(iFile, iEvent);
      vNofHitsPerPoint[iE].resize(nPoints, 0);
    }
  }

  for (int iH = 0; iH < nHits; ++iH) {
    const auto* pHit = dynamic_cast<const CbmTrdHit*>(fpHits->At(iH));
    LOG_IF(fatal, !pHit) << fName << ": hit with iH = " << iH << " is not an CbmTrdHit (dynamic cast failed)";

    // *************************
    // ** Reconstructed hit QA

    // Hit station geometry info
    int iSt = fpDetInterface->GetTrackingStationIndex(pHit);
    LOG_IF(fatal, iSt < 0 || iSt >= nSt) << fName << ": index of station (" << iSt << ") is out of range "
                                         << "for hit with id = " << iH;

    //int hitType = pHit->GetClassType();  // TRD-1D, TRD-2D

    // Hit position
    double xHit = pHit->GetX();
    double yHit = pHit->GetY();
    double zHit = pHit->GetZ();

    double dxHit = pHit->GetDx();
    double dyHit = pHit->GetDy();

    // Hit time
    double tHit  = pHit->GetTime();
    double dtHit = pHit->GetTimeError();

    fvph_hit_ypos_vs_xpos[iSt]->Fill(xHit, yHit);
    fvph_hit_xpos_vs_zpos[iSt]->Fill(zHit, xHit);
    fvph_hit_ypos_vs_zpos[iSt]->Fill(zHit, yHit);

    fvph_hit_station_delta_z[iSt]->Fill(zHit - fpDetInterface->GetZ(iSt));

    fvph_hit_dx[iSt]->Fill(dxHit);
    fvph_hit_dy[iSt]->Fill(dyHit);
    fvph_hit_dt[iSt]->Fill(dtHit);

    // **********************
    // ** MC information QA

    if (IsMCUsed()) {
      CbmMatch* pHitMatch = dynamic_cast<CbmMatch*>(fpHitMatches->At(iH));
      LOG_IF(fatal, !pHitMatch) << fName << ": match object not found for hit ID " << iH;

      // Evaluate number of hits per one MC point
      int nMCpoints = 0;  // Number of MC points for this hit

      for (int iLink = 0; iLink < pHitMatch->GetNofLinks(); ++iLink) {
        const auto& link = pHitMatch->GetLink(iLink);

        int iP = link.GetIndex();  // Index of MC point

        // Skip noisy links
        if (iP < 0) { continue; }

        ++nMCpoints;

        int iE = fpMCEventList->GetEventIndex(link);

        LOG_IF(fatal, iE < 0 || iE >= nMCevents) << fName << ": id of MC event is out of range (hit id = " << iH
                                                 << ", link id = " << iLink << ", event id = " << iE << ')';

        LOG_IF(fatal, iP >= (int) vNofHitsPerPoint[iE].size())
          << fName << ": index of MC point is out of range (hit id = " << iH << ", link id = " << iLink
          << ", event id = " << iE << ", point id = " << iP << ')';
        vNofHitsPerPoint[iE][iP]++;
      }

      fvph_n_points_per_hit[iSt]->Fill(nMCpoints);

      if (nMCpoints != 1) { continue; }  // ??

      // The best link to in the match (probably, the cut on nMCpoints is meaningless)
      const auto& bestPointLink = pHitMatch->GetMatchedLink();

      // Skip noisy links
      if (bestPointLink.GetIndex() < 0) { continue; }

      // Point matched by the best link
      const auto* pMCPoint = dynamic_cast<const CbmTrdPoint*>(fpMCPoints->Get(bestPointLink));
      LOG_IF(fatal, !pMCPoint) << fName << ": MC point object does not exist for hit " << iH;

      // MC track for this point
      CbmLink bestTrackLink = bestPointLink;
      bestTrackLink.SetIndex(pMCPoint->GetTrackID());
      const auto* pMCTrack = dynamic_cast<const CbmMCTrack*>(fpMCTracks->Get(bestTrackLink));
      LOG_IF(fatal, !pMCTrack) << fName << ": MC track object does not exist for hit " << iH << " and link: ";

      double t0MC = fpMCEventList->GetEventTime(bestPointLink);
      LOG_IF(fatal, t0MC < 0) << fName << ": MC time zero is lower then 0 ns: " << t0MC;

      // ----- MC point properties
      //
      double mass = pMCTrack->GetMass();
      //double charge = pMCTrack->GetCharge();
      //double pdg    = pMCTrack->GetPdgCode();

      // Entrance position and time
      double xMC = pMCPoint->GetX();
      double yMC = pMCPoint->GetY();
      double zMC = pMCPoint->GetZ();
      double tMC = pMCPoint->GetTime() + t0MC;

      // MC point entrance momenta
      double pxMC = pMCPoint->GetPx();
      double pyMC = pMCPoint->GetPy();
      double pzMC = pMCPoint->GetPz();
      double pMC  = sqrt(pxMC * pxMC + pyMC * pyMC + pzMC * pzMC);

      // MC point exit momenta
      // double pxMCo = pMCPoint->GetPxOut();
      // double pyMCo = pMCPoint->GetPyOut();
      // double pzMCo = pMCPoint->GetPzOut();
      // double pMCo  = sqrt(pxMCo * pxMCo + pyMCo * pyMCo + pzMCo * pzMCo);

      // Position and time shifted to z-coordinate of the hit
      double shiftZ = zHit - zMC;  // Difference between hit and point z positions
      double xMCs   = xMC + shiftZ * pxMC / pzMC;
      double yMCs   = yMC + shiftZ * pyMC / pzMC;
      double tMCs   = tMC + shiftZ / (pzMC * phys::kSpeedOfLight) * sqrt(mass * mass + pMC * pMC);

      // Residuals
      double xRes = xHit - xMCs;
      double yRes = yHit - yMCs;
      double tRes = tHit - tMCs;

      // ------ Cuts

      if (std::fabs(pMCPoint->GetPzOut()) < fMinMomentum) { continue; }  // CUT ON MINIMUM MOMENTUM
      //if (pMCo < cuts::kMinP) { continue; }  // Momentum threshold


      fvph_point_ypos_vs_xpos[iSt]->Fill(xMC, yMC);
      fvph_point_xpos_vs_zpos[iSt]->Fill(zMC, xMC);
      fvph_point_ypos_vs_zpos[iSt]->Fill(zMC, yMC);

      fvph_point_hit_delta_z[iSt]->Fill(shiftZ);

      fvph_res_x[iSt]->Fill(xRes);
      fvph_res_y[iSt]->Fill(yRes);
      fvph_res_t[iSt]->Fill(tRes);

      fvph_pull_x[iSt]->Fill(xRes / dxHit);
      fvph_pull_y[iSt]->Fill(yRes / dyHit);
      fvph_pull_t[iSt]->Fill(tRes / dtHit);

      fvph_res_x_vs_x[iSt]->Fill(xHit, xRes);
      fvph_res_y_vs_y[iSt]->Fill(yHit, yRes);
      fvph_res_t_vs_t[iSt]->Fill(tHit, tRes);

      fvph_pull_x_vs_x[iSt]->Fill(xHit, xRes / dxHit);
      fvph_pull_y_vs_y[iSt]->Fill(yHit, yRes / dyHit);
      fvph_pull_t_vs_t[iSt]->Fill(tHit, tRes / dtHit);
    }
  }  // loop over hits: end

  // Fill hit reconstruction efficiencies
  if (IsMCUsed()) {
    for (int iE = 0; iE < nMCevents; ++iE) {
      int iFile   = fpMCEventList->GetFileIdByIndex(iE);
      int iEvent  = fpMCEventList->GetEventIdByIndex(iE);
      int nPoints = fpMCPoints->Size(iFile, iEvent);

      for (int iP = 0; iP < nPoints; ++iP) {
        const auto* pMCPoint = dynamic_cast<const CbmTrdPoint*>(fpMCPoints->Get(iFile, iEvent, iP));
        LOG_IF(fatal, !pMCPoint) << fName << ": MC point does not exist for iFile = " << iFile
                                 << ", iEvent = " << iEvent << ", iP = " << iP;
        int address = pMCPoint->GetDetectorID();
        int iSt     = fpDetInterface->GetTrackingStationIndex(address);
        LOG_IF(fatal, iSt < 0 || iSt >= nSt)
          << fName << ": MC point for FEI = " << iFile << ", " << iEvent << ", " << iP << " and address " << address
          << " has wrong station index: iSt = " << iSt;

        double xMC = pMCPoint->GetXIn();
        double yMC = pMCPoint->GetYIn();
        double rMC = sqrt(xMC * xMC + yMC * yMC);

        // Conditions under which point is accounted as reconstructed: point
        bool ifPointHasHits = (vNofHitsPerPoint[iE][iP] > 0);

        fvpe_reco_eff_vs_xy[iSt]->Fill(xMC, yMC, ifPointHasHits);
        fvpe_reco_eff_vs_r[iSt]->Fill(rMC, ifPointHasHits);

      }  // loop over MC-points: end
    }    // loop over MC-events: end
  }      // MC is used: end
}

// ---------------------------------------------------------------------------------------------------------------------
//
InitStatus CbmCaInputQaTrd::InitDataBranches()
{
  // STS tracking detector interface
  fpDetInterface = CbmTrdTrackingInterface::Instance();

  LOG_IF(fatal, !fpDetInterface) << "\033[1;31m" << fName << ": tracking detector interface is undefined\033[0m";

  // FairRootManager
  auto* pFairRootManager = FairRootManager::Instance();
  LOG_IF(fatal, !pFairRootManager) << "\033[1;31m" << fName << ": FairRootManager instance is a null pointer\033[0m";

  // Time-slice
  fpTimeSlice = dynamic_cast<CbmTimeSlice*>(pFairRootManager->GetObject("TimeSlice."));
  LOG_IF(fatal, !fpTimeSlice) << "\033[1;31m" << fName << ": time-slice branch is not found\033[0m";

  // ----- Reconstructed data-branches initialization

  // Hits container
  fpHits = dynamic_cast<TClonesArray*>(pFairRootManager->GetObject("TrdHit"));
  LOG_IF(fatal, !fpHits) << "\033[1;31m" << fName << ": container of reconstructed hits in TRD is not found\033[0m";

  // ----- MC-information branches initialization
  if (IsMCUsed()) {
    // MC manager
    fpMCDataManager = dynamic_cast<CbmMCDataManager*>(pFairRootManager->GetObject("MCDataManager"));
    LOG_IF(fatal, !fpMCDataManager) << "\033[1;31m" << fName << ": MC data manager branch is not found\033[0m";

    // MC event list
    fpMCEventList = dynamic_cast<CbmMCEventList*>(pFairRootManager->GetObject("MCEventList."));
    LOG_IF(fatal, !fpMCEventList) << "\033[1;31m" << fName << ": MC event list branch is not found\033[0m";

    // MC tracks
    fpMCTracks = fpMCDataManager->InitBranch("MCTrack");
    LOG_IF(fatal, !fpMCTracks) << "\033[1;31m" << fName << ": MC track branch is not found\033[0m";

    // MC points
    fpMCPoints = fpMCDataManager->InitBranch("TrdPoint");
    LOG_IF(fatal, !fpMCTracks) << "\033[1;31m" << fName << ": MC point branch is not found for TRD\033[0m";

    // Hit matches
    fpHitMatches = dynamic_cast<TClonesArray*>(pFairRootManager->GetObject("TrdHitMatch"));
    LOG_IF(fatal, !fpHitMatches) << "\033[1;31m]" << fName << ": hit match branch is not found for TRD\033[0m";
  }

  return kSUCCESS;
}

// ---------------------------------------------------------------------------------------------------------------------
//
InitStatus CbmCaInputQaTrd::InitCanvases()
{
  gStyle->SetOptFit(1);
  int nSt = fpDetInterface->GetNtrackingStations();

  // ********************
  // ** Drawing options

  // Contours
  constexpr auto contColor = kOrange + 7;
  constexpr auto contWidth = 2;  // Line width
  constexpr auto contStyle = 2;  // Line style
  constexpr auto contFill  = 0;  // Fill style

  // *********************************
  // ** Hit occupancies

  // ** Occupancy in XY plane
  auto* pc_hit_ypos_vs_xpos = MakeQaObject<CbmQaCanvas>("hit_ypos_vs_xpos", "", 1600, 800);
  pc_hit_ypos_vs_xpos->Divide2D(nSt);
  for (int iSt = 0; iSt < nSt; ++iSt) {
    pc_hit_ypos_vs_xpos->cd(iSt + 1);
    fvph_hit_ypos_vs_xpos[iSt]->DrawCopy("colz", "");


    // Reference size of the station saved to the detector interface
    auto* pInnerCircle = new TEllipse(0., 0., fpDetInterface->GetRmin(iSt));
    pInnerCircle->SetLineWidth(contWidth);
    pInnerCircle->SetLineStyle(contStyle);
    pInnerCircle->SetLineColor(contColor);
    pInnerCircle->SetFillStyle(contFill);
    pInnerCircle->Draw("SAME");

    auto* pOuterCircle = new TEllipse(0., 0., fpDetInterface->GetRmax(iSt));
    pOuterCircle->SetLineWidth(contWidth);
    pOuterCircle->SetLineStyle(contStyle);
    pOuterCircle->SetLineColor(contColor);
    pOuterCircle->SetFillStyle(contFill);
    pOuterCircle->Draw("SAME");

    // Square boarders of station, which are used for the field approximation
    double stXmin = +fpDetInterface->GetXmax(iSt);
    double stXmax = -fpDetInterface->GetXmax(iSt);
    double stYmin = -fpDetInterface->GetYmax(iSt);
    double stYmax = +fpDetInterface->GetYmax(iSt);

    auto* pBox = new TBox(stXmin, stYmin, stXmax, stYmax);
    pBox->SetLineWidth(contWidth);
    pBox->SetLineStyle(contStyle);
    pBox->SetLineColor(contColor);
    pBox->SetFillStyle(contFill);
    pBox->Draw("SAME");
  }

  // ----- Occupancy projections
  auto* pc_hit_xpos = MakeQaObject<CbmQaCanvas>("hit_xpos", "", 1400, 800);
  pc_hit_xpos->Divide2D(nSt);
  for (int iSt = 0; iSt < nSt; ++iSt) {
    pc_hit_xpos->cd(iSt + 1);
    auto* phProj = fvph_hit_ypos_vs_xpos[iSt]->ProjectionY();
    phProj->DrawCopy();
  }

  auto* pc_hit_ypos = MakeQaObject<CbmQaCanvas>("hit_ypos", "", 1400, 800);
  pc_hit_ypos->Divide2D(nSt);
  for (int iSt = 0; iSt < nSt; ++iSt) {
    pc_hit_ypos->cd(iSt + 1);
    auto* phProj = fvph_hit_ypos_vs_xpos[iSt]->ProjectionX();
    phProj->DrawCopy();
  }

  // ************
  // ************

  if (IsMCUsed()) {

    // **********************
    // ** Point occupancies

    // ** Occupancy in XY plane
    auto* pc_point_ypos_vs_xpos = MakeQaObject<CbmQaCanvas>("point_ypos_vs_xpos", "", 1600, 800);
    pc_point_ypos_vs_xpos->Divide2D(nSt);
    for (int iSt = 0; iSt < nSt; ++iSt) {
      pc_point_ypos_vs_xpos->cd(iSt + 1);
      fvph_point_ypos_vs_xpos[iSt]->DrawCopy("colz", "");

      // Reference size of the station saved to the detector interface
      auto* pInnerCircle = new TEllipse(0., 0., fpDetInterface->GetRmin(iSt));
      pInnerCircle->SetLineWidth(contWidth);
      pInnerCircle->SetLineStyle(contStyle);
      pInnerCircle->SetLineColor(contColor);
      pInnerCircle->SetFillStyle(contFill);
      pInnerCircle->Draw("SAME");

      auto* pOuterCircle = new TEllipse(0., 0., fpDetInterface->GetRmax(iSt));
      pOuterCircle->SetLineWidth(contWidth);
      pOuterCircle->SetLineStyle(contStyle);
      pOuterCircle->SetLineColor(contColor);
      pOuterCircle->SetFillStyle(contFill);
      pOuterCircle->Draw("SAME");

      // Square boarders of station, which are used for the field approximation
      double stXmin = +fpDetInterface->GetXmax(iSt);
      double stXmax = -fpDetInterface->GetXmax(iSt);
      double stYmin = -fpDetInterface->GetYmax(iSt);
      double stYmax = +fpDetInterface->GetYmax(iSt);

      auto* pBox = new TBox(stXmin, stYmin, stXmax, stYmax);
      pBox->SetLineWidth(contWidth);
      pBox->SetLineStyle(contStyle);
      pBox->SetLineColor(contColor);
      pBox->SetFillStyle(contFill);
      pBox->Draw("SAME");
    }

    // **************
    // ** Residuals

    // x-coordinate
    auto* pc_res_x = MakeQaObject<CbmQaCanvas>("res_x", "Residuals for x coordinate", 1600, 800);
    pc_res_x->Divide2D(nSt);
    for (int iSt = 0; iSt < nSt; ++iSt) {
      pc_res_x->cd(iSt + 1);
      fvph_res_x[iSt]->DrawCopy("", "");
    }

    // y-coordinate
    auto* pc_res_y = MakeQaObject<CbmQaCanvas>("res_y", "Residuals for y coordinate", 1600, 800);
    pc_res_y->Divide2D(nSt);
    for (int iSt = 0; iSt < nSt; ++iSt) {
      pc_res_y->cd(iSt + 1);
      fvph_res_y[iSt]->DrawCopy("", "");
    }

    // time
    auto* pc_res_t = MakeQaObject<CbmQaCanvas>("res_t", "Residuals for time", 1600, 800);
    pc_res_t->Divide2D(nSt);
    for (int iSt = 0; iSt < nSt; ++iSt) {
      pc_res_t->cd(iSt + 1);
      fvph_res_t[iSt]->DrawCopy("", "");
    }


    // **********
    // ** Pulls

    // x-coordinate
    auto* pc_pull_x = MakeQaObject<CbmQaCanvas>("pull_x", "Pulls for x coordinate", 1600, 800);
    pc_pull_x->Divide2D(nSt);
    for (int iSt = 0; iSt < nSt; ++iSt) {
      pc_pull_x->cd(iSt + 1);
      fvph_pull_x[iSt]->DrawCopy("", "");
    }

    // y-coordinate
    auto* pc_pull_y = MakeQaObject<CbmQaCanvas>("pull_y", "Pulls for y coordinate", 1600, 800);
    pc_pull_y->Divide2D(nSt);
    for (int iSt = 0; iSt < nSt; ++iSt) {
      pc_pull_y->cd(iSt + 1);
      fvph_pull_y[iSt]->DrawCopy("", "");
    }

    // time
    auto* pc_pull_t = MakeQaObject<CbmQaCanvas>("pull_t", "Pulls for time", 1600, 800);
    pc_pull_t->Divide2D(nSt);
    for (int iSt = 0; iSt < nSt; ++iSt) {
      pc_pull_t->cd(iSt + 1);
      fvph_pull_t[iSt]->DrawCopy("", "");
    }


    // ************************************
    // ** Hit reconstruction efficiencies

    auto* pc_reco_eff_vs_r =
      MakeQaObject<CbmQaCanvas>("reco_eff_vs_r", "Hit efficiencies wrt distance from center", 1600, 800);
    pc_reco_eff_vs_r->Divide2D(nSt);
    for (int iSt = 0; iSt < nSt; ++iSt) {
      pc_reco_eff_vs_r->cd(iSt + 1);
      fvpe_reco_eff_vs_r[iSt]->DrawCopy("AP", "");
    }

    auto* pc_reco_eff_vs_xy = MakeQaObject<CbmQaCanvas>("reco_eff_vs_xy", "Hit efficiencies wrt x and y", 1600, 800);
    pc_reco_eff_vs_xy->Divide2D(nSt);
    for (int iSt = 0; iSt < nSt; ++iSt) {
      pc_reco_eff_vs_xy->cd(iSt + 1);
      fvpe_reco_eff_vs_xy[iSt]->DrawCopy("colz", "");
    }
  }

  return kSUCCESS;
}

// ---------------------------------------------------------------------------------------------------------------------
//
InitStatus CbmCaInputQaTrd::InitHistograms()
{
  int nSt = fpDetInterface->GetNtrackingStations();

  // ----- Histograms initialization
  fvph_hit_ypos_vs_xpos.resize(nSt, nullptr);
  fvph_hit_ypos_vs_zpos.resize(nSt, nullptr);
  fvph_hit_xpos_vs_zpos.resize(nSt, nullptr);

  fvph_hit_station_delta_z.resize(nSt);

  fvph_hit_dx.resize(nSt, nullptr);
  fvph_hit_dy.resize(nSt, nullptr);
  fvph_hit_dt.resize(nSt, nullptr);

  if (IsMCUsed()) {
    fvph_n_points_per_hit.resize(nSt, nullptr);
    fvph_point_ypos_vs_xpos.resize(nSt, nullptr);
    fvph_point_xpos_vs_zpos.resize(nSt, nullptr);
    fvph_point_ypos_vs_zpos.resize(nSt, nullptr);
    fvph_point_hit_delta_z.resize(nSt, nullptr);
    fvph_res_x.resize(nSt, nullptr);
    fvph_res_y.resize(nSt, nullptr);
    fvph_res_t.resize(nSt, nullptr);
    fvph_pull_x.resize(nSt, nullptr);
    fvph_pull_y.resize(nSt, nullptr);
    fvph_pull_t.resize(nSt, nullptr);
    fvph_res_x_vs_x.resize(nSt, nullptr);
    fvph_res_y_vs_y.resize(nSt, nullptr);
    fvph_res_t_vs_t.resize(nSt, nullptr);
    fvph_pull_x_vs_x.resize(nSt, nullptr);
    fvph_pull_y_vs_y.resize(nSt, nullptr);
    fvph_pull_t_vs_t.resize(nSt, nullptr);
    fvpe_reco_eff_vs_xy.resize(nSt, nullptr);
    fvpe_reco_eff_vs_r.resize(nSt, nullptr);
  }

  for (int iSt = 0; iSt < nSt; ++iSt) {
    TString nsuff        = Form("_st%d", iSt);
    TString tsuff        = Form(" in TRD station %d", iSt);
    TString npref        = Form("station%d/", iSt);
    TString xyBinningTag = Form("xy_station%d", iSt);

    TString sN = "";
    TString sT = "";

    double xMax = fpDetInterface->GetXmax(iSt);
    double yMax = fpDetInterface->GetYmax(iSt);
    double xRes = 0.05;  // [cm]
    double yRes = 0.20;  // [cm]
    int nBinsX  = static_cast<int>(2 * xMax / xRes);
    int nBinsY  = static_cast<int>(2 * yMax / yRes);

    // Hit occupancy
    sN                         = npref + "hit_ypos_vs_xpos" + nsuff + ";" + xyBinningTag;
    sT                         = (TString) "Hit occupancy in xy-Plane" + tsuff + ";x_{hit} [cm];y_{hit} [cm]";
    fvph_hit_ypos_vs_xpos[iSt] = MakeQaObject<TH2F>(sN, sT, nBinsX, -xMax, xMax, nBinsY, -yMax, yMax);

    sN                         = npref + "hit_xpos_vs_zpos" + nsuff;
    sT                         = (TString) "Hit occupancy in xz-Plane" + tsuff + ";z_{hit} [cm];x_{hit} [cm]";
    fvph_hit_xpos_vs_zpos[iSt] = MakeQaObject<TH2F>(sN, sT, fNbinsZ, fvRHitZ[0], fvRHitZ[1], nBinsX, -xMax, xMax);

    sN                         = npref + "hit_ypos_vs_zpos" + nsuff;
    sT                         = (TString) "Hit occupancy in yz-plane" + tsuff + ";z_{hit} [cm];y_{hit} [cm]";
    fvph_hit_ypos_vs_zpos[iSt] = MakeQaObject<TH2F>(sN, sT, fNbinsZ, fvRHitZ[0], fvRHitZ[1], nBinsY, -yMax, yMax);

    // Hit errors
    sN               = npref + "hit_dx" + nsuff;
    sT               = (TString) "Hit position error along x-axis" + tsuff + ";dx_{hit} [cm]";
    fvph_hit_dx[iSt] = MakeQaObject<TH1F>(sN, sT, fNbins, fvRHitDx[0], fvRHitDx[1]);

    sN               = npref + "hit_dy" + nsuff;
    sT               = (TString) "Hit position error along y-axis" + tsuff + ";dy_{hit} [cm]";
    fvph_hit_dy[iSt] = MakeQaObject<TH1F>(sN, sT, fNbins, fvRHitDy[0], fvRHitDy[1]);

    sN               = npref + "hit_dt" + nsuff;
    sT               = (TString) "Hit time error" + tsuff + ";dt_{hit} [ns]";
    fvph_hit_dt[iSt] = MakeQaObject<TH1F>(sN, sT, fNbins, fvRHitDt[0], fvRHitDt[1]);

    sN = npref + "hit_station_delta_z" + nsuff;
    sT = (TString) "Different between hit and station z-positions" + tsuff + ";z_{hit} - z_{st} [cm]";
    fvph_hit_station_delta_z[iSt] = MakeQaObject<TH1F>(sN, sT, fNbins, -5., 5);

    // ----- Initialize histograms, which are use MC-information
    if (IsMCUsed()) {
      sN                         = npref + "n_points_per_hit" + nsuff;
      sT                         = (TString) "Number of points per hit" + tsuff + ";N_{point}/hit";
      fvph_n_points_per_hit[iSt] = MakeQaObject<TH1F>(sN, sT, 10, -0.5, 9.5);

      // Point occupancy
      sN                           = npref + "point_ypos_vs_xpos" + nsuff + ";" + xyBinningTag;
      sT = (TString) "Point occupancy in XY plane" + tsuff + ";x_{MC} [cm];y_{MC} [cm]";
      fvph_point_ypos_vs_xpos[iSt] = MakeQaObject<TH2F>(sN, sT, nBinsX, -xMax, xMax, nBinsY, -yMax, yMax);

      sN                           = npref + "point_xpos_vs_zpos" + nsuff;
      sT = (TString) "Point Occupancy in XZ plane" + tsuff + ";z_{MC} [cm];x_{MC} [cm]";
      fvph_point_xpos_vs_zpos[iSt] = MakeQaObject<TH2F>(sN, sT, fNbinsZ, fvRHitZ[0], fvRHitZ[1], nBinsX, -xMax, xMax);

      sN                           = npref + "point_ypos_vs_zpos" + nsuff;
      sT = (TString) "Point Occupancy in YZ Plane" + tsuff + ";z_{MC} [cm];y_{MC} [cm]";
      fvph_point_ypos_vs_zpos[iSt] = MakeQaObject<TH2F>(sN, sT, fNbinsZ, fvRHitZ[0], fvRHitZ[1], nBinsY, -yMax, yMax);

      // Difference between MC input z and hit z coordinates
      sN = npref + "point_hit_delta_z" + nsuff;
      sT = (TString) "Distance between point and hit along z axis" + tsuff + ";z_{reco} - z_{MC} [cm]";
      fvph_point_hit_delta_z[iSt] = MakeQaObject<TH1F>(sN, sT, fNbins, -0.04, 0.04);

      sN              = npref + "res_x" + nsuff;
      sT              = (TString) "Residuals for x-coordinate" + tsuff + ";x_{reco} - x_{MC} [cm]";
      fvph_res_x[iSt] = MakeQaObject<TH1F>(sN, sT, fNbins, fvRResX[0], fvRResX[1]);

      sN              = npref + "res_y" + nsuff;
      sT              = (TString) "Residuals for y-coordinate" + tsuff + ";y_{reco} - y_{MC} [cm]";
      fvph_res_y[iSt] = MakeQaObject<TH1F>(sN, sT, fNbins, fvRResY[0], fvRResY[1]);

      sN              = npref + "res_t" + nsuff;
      sT              = (TString) "Residuals for time" + tsuff + ";t_{reco} - t_{MC} [ns]";
      fvph_res_t[iSt] = MakeQaObject<TH1F>(sN, sT, fNbins, fvRResT[0], fvRResT[1]);

      sN               = npref + "pull_x" + nsuff;
      sT               = (TString) "Pulls for x-coordinate" + tsuff + ";(x_{reco} - x_{MC}) / #sigma_{x}^{reco}";
      fvph_pull_x[iSt] = MakeQaObject<TH1F>(sN, sT, fNbins, fvRPullX[0], fvRPullX[1]);

      sN               = npref + "pull_y" + nsuff;
      sT               = (TString) "Pulls for y-coordinate" + tsuff + ";(y_{reco} - y_{MC}) / #sigma_{y}^{reco}";
      fvph_pull_y[iSt] = MakeQaObject<TH1F>(sN, sT, fNbins, fvRPullY[0], fvRPullY[1]);

      sN               = npref + "pull_t" + nsuff;
      sT               = (TString) "Pulls for time" + tsuff + ";(t_{reco} - t_{MC}) / #sigma_{t}^{reco}";
      fvph_pull_t[iSt] = MakeQaObject<TH1F>(sN, sT, fNbins, fvRPullT[0], fvRPullT[1]);

      sN                   = npref + "res_x_vs_x" + nsuff;
      sT                   = (TString) "Residuals for x-coordinate" + tsuff + ";x_{reco} [cm];x_{reco} - x_{MC} [cm]";
      fvph_res_x_vs_x[iSt] = MakeQaObject<TH2F>(sN, sT, fNbins, 0, 0, fNbins, fvRResX[0], fvRResX[1]);

      sN                   = npref + "res_y_vs_y" + nsuff;
      sT                   = (TString) "Residuals for y-coordinate" + tsuff + ";y_{reco} [cm];y_{reco} - y_{MC} [cm]";
      fvph_res_y_vs_y[iSt] = MakeQaObject<TH2F>(sN, sT, fNbins, 0, 0, fNbins, fvRResY[0], fvRResY[1]);

      sN                   = npref + "res_t_vs_t" + nsuff;
      sT                   = (TString) "Residuals for time" + tsuff + "t_{reco} [ns];t_{reco} - t_{MC} [ns]";
      fvph_res_t_vs_t[iSt] = MakeQaObject<TH2F>(sN, sT, fNbins, 0, 0, fNbins, fvRResT[0], fvRResT[1]);

      sN = npref + "pull_x_vs_x" + nsuff;
      sT = (TString) "Pulls for x-coordinate" + tsuff + "x_{reco} [cm];(x_{reco} - x_{MC}) / #sigma_{x}^{reco}";
      fvph_pull_x_vs_x[iSt] = MakeQaObject<TH2F>(sN, sT, fNbins, 0, 0, fNbins, fvRPullX[0], fvRPullX[1]);

      sN = npref + "pull_y_vs_y" + nsuff;
      sT = (TString) "Pulls for y-coordinate" + tsuff + ";y_{reco} [cm];(y_{reco} - y_{MC}) / #sigma_{y}^{reco}";
      fvph_pull_y_vs_y[iSt] = MakeQaObject<TH2F>(sN, sT, fNbins, 0, 0, fNbins, fvRPullY[0], fvRPullY[1]);

      sN                    = npref + "pull_t_vs_t" + nsuff;
      sT = (TString) "Pulls for time" + tsuff + ";t_{reco} [ns];(t_{reco} - t_{MC}) / #sigma_{t}^{reco}";
      fvph_pull_t_vs_t[iSt] = MakeQaObject<TH2F>(sN, sT, fNbins, 0, 0, fNbins, fvRPullT[0], fvRPullT[1]);

      sN                       = npref + "reco_eff_vs_xy" + nsuff + ";" + xyBinningTag;
      sT                       = (TString) "Hit rec. efficiency" + tsuff + ";x_{MC} [cm];y_{MC} [cm];#epsilon";
      fvpe_reco_eff_vs_xy[iSt] =
        MakeQaObject<TProfile2D>(sN, sT, fNbins, fvRHitX[0], fvRHitX[1], fNbins, fvRHitY[0], fvRHitY[1]);

      sN                      = npref + "reco_eff_vs_r" + nsuff;
      sT                      = (TString) "Hit rec. efficiency" + tsuff + ";r_{MC} [cm];#epsilon";
      fvpe_reco_eff_vs_r[iSt] = MakeQaObject<TProfile>(sN, sT, 100, 0, 100, 0., 1.);
    }
  }
  return kSUCCESS;
}
