/* Copyright (C) 2022 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Volker Friese [committer] */


#ifndef CBMTASKUNPACK_H
#define CBMTASKUNPACK_H 1

#include "CbmDefs.h"
#include "CbmDigiTimeslice.h"

// Hide the Monitor class from ROOT to not confuse it
#if !defined(__CLING__) && !defined(__ROOTCLING__)
#include "Monitor.hpp"
#else
namespace cbm
{
  class Monitor;
}
#endif

#include <FairTask.h>

#include <sstream>
#include <vector>

#include "EventBuilder.h"
#include "Unpack.h"

class CbmDigiManager;
class CbmSourceTs;


/** @class CbmTaskUnpack
 ** @brief Task class for associating digis to events
 ** @author Volker Friese <v.friese@gsi.de>
 ** @since 15.11.2021
 **
 ** Creates objects of class CbmDigiEvent and fills them with digi objects,
 ** using the algorithm EventBuilder.
 **
 ** TOFO: The current implementation is for STS only and with a dummy trigger list
 ** just to establish the framework integration of algorithm and data interfaces.
 **/
class CbmTaskUnpack : public FairTask {


public:
  /** @brief Constructor **/
  CbmTaskUnpack();


  /** @brief Copy constructor (disabled) **/
  CbmTaskUnpack(const CbmTaskUnpack&) = delete;


  /** @brief Destructor **/
  virtual ~CbmTaskUnpack();


  /** @brief Task execution **/
  virtual void Exec(Option_t* opt);


  /** @brief Finish timeslice **/
  virtual void Finish();


  /** @brief Assignment operator (disabled) **/
  CbmTaskUnpack& operator=(const CbmTaskUnpack&) = delete;


  /** @brief Set the monitor object **/
  void SetMonitor(cbm::Monitor* monitor) { fMonitor = monitor; }


private:  // methods
  /** @brief Task initialisation **/
  virtual InitStatus Init();


  /** @brief Initialisation of address maps for Trd **/
  cbm::algo::trd::ReadoutConfig InitTrdReadoutConfig();


  /** @brief Initialisation of address maps for Trd2d **/
  cbm::algo::trd2d::ReadoutConfig InitTrd2dReadoutConfig();

private:  // members
  CbmSourceTs* fSource   = nullptr;
  cbm::Monitor* fMonitor = nullptr;
  std::string fHostname;

  /** @brief Unpacker algorithm **/
  cbm::algo::Unpack fUnpack;

  size_t fNumTs                = 0;
  size_t fNumMs                = 0;
  size_t fNumBytes             = 0;
  size_t fNumDigis             = 0;
  double fTime                 = 0.;
  CbmDigiTimeslice* fTimeslice = nullptr;  ///< Output data

  ClassDef(CbmTaskUnpack, 3);
};

#endif /* CBMTASKUNPACK_H */
